# Generated by Django 4.2.7 on 2023-12-04 22:19

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        (
            "edc_visit_tracking",
            "0004_subjectvisit_subjectvisitmissedreasons_extra_value_and_more",
        ),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="subjectvisit",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
            },
        ),
        migrations.AlterModelOptions(
            name="subjectvisitmissed",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "verbose_name": "Missed Visit Report",
                "verbose_name_plural": "Missed Visit Report",
            },
        ),
        migrations.AlterModelOptions(
            name="subjectvisitmissedreasons",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "verbose_name": "Subject Missed Visit Reasons",
                "verbose_name_plural": "Subject Missed Visit Reasons",
            },
        ),
        migrations.AlterModelOptions(
            name="visitreasons",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "verbose_name": "Visit Reasons",
                "verbose_name_plural": "Visit Reasons",
            },
        ),
        migrations.RemoveIndex(
            model_name="subjectvisitmissed",
            name="edc_visit_t_subject_8fc4e6_idx",
        ),
        migrations.RemoveIndex(
            model_name="subjectvisitmissedreasons",
            name="edc_visit_t_id_a0ee54_idx",
        ),
        migrations.RemoveIndex(
            model_name="visitreasons",
            name="edc_visit_t_id_01998c_idx",
        ),
        migrations.AddField(
            model_name="historicalsubjectvisit",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="historicalsubjectvisit",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="historicalsubjectvisitmissed",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="historicalsubjectvisitmissed",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="subjectvisit",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="subjectvisit",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="subjectvisitmissed",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="subjectvisitmissed",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AlterField(
            model_name="subjectvisitmissedreasons",
            name="display_index",
            field=models.IntegerField(
                default=0,
                help_text="Index to control display order if not alphabetical, not required",
                verbose_name="display index",
            ),
        ),
        migrations.AlterField(
            model_name="subjectvisitmissedreasons",
            name="display_name",
            field=models.CharField(
                help_text="(suggest 40 characters max.)",
                max_length=250,
                unique=True,
                verbose_name="Name",
            ),
        ),
        migrations.AlterField(
            model_name="subjectvisitmissedreasons",
            name="name",
            field=models.CharField(
                help_text="This is the stored value, required",
                max_length=250,
                unique=True,
                verbose_name="Stored value",
            ),
        ),
        migrations.AlterField(
            model_name="visitreasons",
            name="display_index",
            field=models.IntegerField(
                default=0,
                help_text="Index to control display order if not alphabetical, not required",
                verbose_name="display index",
            ),
        ),
        migrations.AlterField(
            model_name="visitreasons",
            name="display_name",
            field=models.CharField(
                help_text="(suggest 40 characters max.)",
                max_length=250,
                unique=True,
                verbose_name="Name",
            ),
        ),
        migrations.AlterField(
            model_name="visitreasons",
            name="name",
            field=models.CharField(
                help_text="This is the stored value, required",
                max_length=250,
                unique=True,
                verbose_name="Stored value",
            ),
        ),
        migrations.AddIndex(
            model_name="subjectvisit",
            index=models.Index(
                fields=["modified", "created"], name="edc_visit_t_modifie_417821_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="subjectvisit",
            index=models.Index(
                fields=["user_modified", "user_created"],
                name="edc_visit_t_user_mo_f84d53_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="subjectvisitmissed",
            index=models.Index(
                fields=["subject_visit", "site"], name="edc_visit_t_subject_34c0b3_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="subjectvisitmissed",
            index=models.Index(
                fields=["modified", "created"], name="edc_visit_t_modifie_8805dd_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="subjectvisitmissed",
            index=models.Index(
                fields=["user_modified", "user_created"],
                name="edc_visit_t_user_mo_35bd52_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="subjectvisitmissedreasons",
            index=models.Index(
                fields=["display_index", "display_name"],
                name="edc_visit_t_display_30249b_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="visitreasons",
            index=models.Index(
                fields=["display_index", "display_name"],
                name="edc_visit_t_display_3470ba_idx",
            ),
        ),
    ]



from django.db import models
from django.utils.translation import gettext_lazy as _
from edc_crf.model_mixins import CrfModelMixin
from edc_model.models import BaseUuidModel

from ..model_mixins import SubjectVisitMissedModelMixin
from .subject_visit import SubjectVisit
from .subject_visit_missed_reasons import SubjectVisitMissedReasons


class SubjectVisitMissed(
    CrfModelMixin,
    SubjectVisitMissedModelMixin,
    BaseUuidModel,
):
    subject_visit = models.OneToOneField(
        SubjectVisit,
        on_delete=models.PROTECT,
        related_name="edc_subject_visit",
    )

    missed_reasons = models.ManyToManyField(
        SubjectVisitMissedReasons, blank=True, related_name="default_missed_reasons"
    )

    class Meta(CrfModelMixin.Meta, BaseUuidModel.Meta):
        verbose_name = _("Missed Visit Report")
        verbose_name_plural = _("Missed Visit Report")
        indexes = CrfModelMixin.Meta.indexes + BaseUuidModel.Meta.indexes



from ..exceptions import RelatedVisitFieldError
from ..utils import get_related_visit_model_cls


def get_related_visit_model_attr(model_cls) -> str:
    """Returns the field name for the visit model foreign key
    or raises.

    Note: may also be the reverse relation. For example, if this is
    an `appointment` model class, the relation is a reverse relation
    where the attr name, given related to SubjectVisit, has a fld
    class attr in `subjectvisit` and not `subject_visit`.

    If the reverse attr is not found, check value of `related_name`
    in the field class, otherwise raise an exception.

    If more than one is found, raise an exception.
    """
    attrs = []
    related_visit_model_cls = get_related_visit_model_cls()
    if related_visit_model_cls._meta.proxy is True:
        related_visit_model_cls = related_visit_model_cls._meta.proxy_for_model
    for fld_cls in model_cls._meta.get_fields():
        if (
            fld_cls.related_model is not None
            and fld_cls.related_model == related_visit_model_cls
        ):
            attrs.append(fld_cls.name)
    if len(attrs) > 1:
        raise RelatedVisitFieldError(
            f"More than one field is related to the visit model. See {model_cls}. "
            f"Got {attrs}. "
        )
    elif len(attrs) == 0:
        raise RelatedVisitFieldError(
            f"{model_cls} has no related visit model. "
            f"Expected the related visit model to be an instance "
            "of `VisitModelMixin`."
        )
    return attrs[0]



from django.db import models
from edc_constants.choices import ALIVE_DEAD_UNKNOWN_NA_MISSED, YES_NO
from edc_constants.constants import ALIVE, NOT_APPLICABLE, YES
from edc_model.validators import date_not_future, datetime_not_future
from edc_model_fields.fields import OtherCharField
from edc_protocol.validators import (
    date_not_before_study_start,
    datetime_not_before_study_start,
)
from edc_utils import get_utcnow

from ...choices import VISIT_REASON_UNSCHEDULED


class VisitModelFieldsMixin(models.Model):
    report_datetime = models.DateTimeField(
        verbose_name="Report date and time",
        validators=[datetime_not_before_study_start, datetime_not_future],
        default=get_utcnow,
        help_text="Date and time of this report",
    )

    reason = models.CharField(
        verbose_name="What is the reason for this report?",
        max_length=25,
    )

    reason_unscheduled = models.CharField(
        verbose_name="If 'unscheduled', provide the reason for " "the unscheduled visit",
        max_length=25,
        choices=VISIT_REASON_UNSCHEDULED,
        default=NOT_APPLICABLE,
    )

    reason_unscheduled_other = OtherCharField(
        verbose_name='If the reason for the unscheduled visit is "other", specify',
        max_length=25,
        blank=True,
        null=True,
    )

    reason_missed = models.CharField(
        verbose_name="If 'missed', provide the reason for the missed visit",
        max_length=35,
        blank=True,
        null=True,
    )

    reason_missed_other = OtherCharField(
        verbose_name='If the reason for the missed visit is "other", specify',
        max_length=25,
        blank=True,
        null=True,
    )

    study_status = models.CharField(
        verbose_name="What is the participant's current study status",
        max_length=50,
        null=True,
    )

    require_crfs = models.CharField(
        max_length=10,
        verbose_name="Are scheduled data being submitted with this visit?",
        choices=YES_NO,
        default=YES,
    )

    info_source = models.CharField(
        verbose_name="What is the MAIN source of this information?", max_length=25
    )

    info_source_other = OtherCharField(
        verbose_name='If "Other" source of information, specify'
    )

    survival_status = models.CharField(
        max_length=10,
        verbose_name="Participant's survival status",
        choices=ALIVE_DEAD_UNKNOWN_NA_MISSED,
        null=True,
        default=ALIVE,
        help_text="If YES, submit Death report",
    )

    last_alive_date = models.DateField(
        verbose_name="Date participant last known alive",
        validators=[date_not_before_study_start, date_not_future],
        null=True,
        blank=True,
        help_text="Will be validated on the Death report",
    )

    comments = models.TextField(
        verbose_name=(
            "Comment if any additional pertinent information " "about the participant"
        ),
        max_length=250,
        blank=True,
        null=True,
    )

    class Meta:
        abstract = True



#!/usr/bin/env python
import sys
from datetime import datetime
from pathlib import Path
from zoneinfo import ZoneInfo

from edc_test_utils import DefaultTestSettings

app_name = "edc_visit_tracking"


base_dir = Path(__file__).parent.parent.parent

project_settings = DefaultTestSettings(
    calling_file=__file__,
    BASE_DIR=base_dir,
    APP_NAME=app_name,
    SITE_ID=1,
    ETC_DIR=str(base_dir / app_name / "tests" / "etc"),
    SILENCED_SYSTEM_CHECKS=[
        "sites.E101",
        "edc_navbar.E002",
        "edc_navbar.E003",
        "edc_consent.E001",
    ],
    SUBJECT_VISIT_MODEL="edc_visit_tracking.subjectvisit",
    EDC_LTFU_MODEL_NAME="edc_ltfu.ltfu",
    EDC_SITES_REGISTER_DEFAULT=True,
    EDC_PROTOCOL_STUDY_OPEN_DATETIME=datetime(
        year=2017, month=1, day=1, tzinfo=ZoneInfo("UTC")
    ),
    EDC_PROTOCOL_STUDY_CLOSE_DATETIME=datetime(
        year=2021, month=1, day=1, tzinfo=ZoneInfo("UTC")
    ),
    INSTALLED_APPS=[
        "django.contrib.admin",
        "django.contrib.auth",
        "django.contrib.contenttypes",
        "django.contrib.messages",
        "django.contrib.sessions",
        "django.contrib.sites",
        "django.contrib.staticfiles",
        "django_crypto_fields.apps.AppConfig",
        "django_revision.apps.AppConfig",
        "multisite",
        "edc_action_item.apps.AppConfig",
        "edc_appointment.apps.AppConfig",
        "edc_auth.apps.AppConfig",
        "edc_data_manager.apps.AppConfig",
        "edc_device.apps.AppConfig",
        "edc_facility.apps.AppConfig",
        "edc_form_runners.apps.AppConfig",
        "edc_identifier.apps.AppConfig",
        "edc_lab.apps.AppConfig",
        "edc_label.apps.AppConfig",
        "edc_locator.apps.AppConfig",
        "edc_metadata.apps.AppConfig",
        "edc_notification.apps.AppConfig",
        "edc_offstudy.apps.AppConfig",
        "edc_registration.apps.AppConfig",
        "edc_sites.apps.AppConfig",
        "edc_timepoint.apps.AppConfig",
        "edc_visit_schedule.apps.AppConfig",
        "edc_visit_tracking.apps.AppConfig",
        "visit_tracking_app.apps.AppConfig",
        "edc_appconfig.apps.AppConfig",
    ],
    add_dashboard_middleware=False,
).settings


for k, v in project_settings.items():
    setattr(sys.modules[__name__], k, v)



from django.db.models.signals import post_delete, post_save
from django.dispatch import receiver
from edc_appointment.constants import ONTIME_APPT
from edc_constants.constants import NOT_APPLICABLE, PATIENT
from edc_metadata import REQUIRED
from edc_metadata.metadata import CrfMetadataGetter
from edc_pharmacy.constants import IN_PROGRESS_APPT

from ..constants import SCHEDULED
from ..model_mixins import SubjectVisitMissedModelMixin


@receiver(post_save, weak=False, dispatch_uid="visit_tracking_check_in_progress_on_post_save")
def visit_tracking_check_in_progress_on_post_save(
    sender, instance, raw, created, using, update_fields, **kwargs  # noqa
):
    """Calls method on the visit tracking instance"""
    if not raw and not update_fields:
        try:
            instance.appointment
        except AttributeError:
            pass
        else:
            try:
                instance.update_appointment_status()
            except AttributeError as e:
                if "update_appointment_status" not in str(e):
                    raise


@receiver(post_delete, weak=False, dispatch_uid="subject_visit_missed_on_post_delete")
def subject_visit_missed_on_post_delete(sender, instance, using, **kwargs) -> None:
    if isinstance(instance, (SubjectVisitMissedModelMixin,)):
        appointment = instance.related_visit.appointment
        # need to remove references and missed visit metadata manually
        getter = CrfMetadataGetter(appointment)
        getter.metadata_objects.filter(model=instance._meta.label_lower).update(
            entry_status=REQUIRED
        )
        # update appointment
        appointment.appt_status = IN_PROGRESS_APPT
        appointment.appt_timing = ONTIME_APPT
        appointment.modified = instance.modified
        appointment.user_modified = instance.user_modified
        appointment.save_base(
            update_fields=["appt_status", "appt_timing", "modified", "user_modified"]
        )
        if appointment.visit_code_sequence == 0:
            # update related visit. Visit code sequence should always be 0
            instance.related_visit.reason = SCHEDULED
            instance.related_visit.reason_unscheduled = NOT_APPLICABLE
            instance.related_visit.info_source = PATIENT
            instance.related_visit.info_source_other = None
            instance.related_visit.comment = None
            instance.related_visit.modified = instance.modified
            instance.related_visit.user_modified = instance.user_modified
            instance.related_visit.save_base(
                update_fields=[
                    "reason",
                    "reason_unscheduled",
                    "info_source",
                    "modified",
                    "user_modified",
                ]
            )



from datetime import datetime
from zoneinfo import ZoneInfo

import time_machine
from dateutil.relativedelta import relativedelta
from django.test import TestCase
from edc_utils import get_utcnow

from edc_visit_tracking.crf_date_validator import (
    CrfDateValidator,
    CrfReportDateAllowanceError,
    CrfReportDateIsFuture,
)

utc_tz = ZoneInfo("UTC")


@time_machine.travel(datetime(2019, 6, 11, 8, 00, tzinfo=utc_tz))
class TestVisitDateValidator(TestCase):
    def test_cls_ok(self):
        dt = get_utcnow()
        CrfDateValidator(report_datetime=dt, visit_report_datetime=dt)

    def test_raises_if_report_datetime_before_visit(self):
        dt = get_utcnow()
        self.assertRaises(
            CrfReportDateAllowanceError,
            CrfDateValidator,
            report_datetime=dt - relativedelta(days=1),
            visit_report_datetime=dt,
        )

    def test_raises_if_report_datetime_is_future(self):
        dt = get_utcnow()
        self.assertRaises(
            CrfReportDateIsFuture,
            CrfDateValidator,
            report_datetime=dt + relativedelta(years=10),
            visit_report_datetime=dt,
        )

    def test_report_datetime_ok(self):
        class MyCrfDateValidator(CrfDateValidator):
            report_datetime_allowance = 3

        visit_report_datetime = get_utcnow() - relativedelta(days=10)
        for days in range(0, 3):
            with self.subTest(days=days):
                try:
                    MyCrfDateValidator(
                        report_datetime=visit_report_datetime + relativedelta(days=days),
                        visit_report_datetime=visit_report_datetime,
                    )
                except CrfReportDateAllowanceError as e:
                    self.fail(f"VisitReportDateAllowanceError unexpectedly raised. Got {e}")

    def test_raises_if_report_datetime(self):
        class MyCrfDateValidator(CrfDateValidator):
            report_datetime_allowance = 3

        visit_report_datetime = get_utcnow() - relativedelta(days=10)
        self.assertRaises(
            CrfReportDateAllowanceError,
            MyCrfDateValidator,
            report_datetime=visit_report_datetime + relativedelta(days=4),
            visit_report_datetime=visit_report_datetime,
        )



from __future__ import annotations

from typing import TYPE_CHECKING

from django.utils.translation import gettext_lazy as _
from edc_appointment.constants import SKIPPED_APPT


class VisitSequenceError(Exception):
    pass


if TYPE_CHECKING:
    from edc_appointment.models import Appointment

    from edc_visit_tracking.model_mixins import VisitModelMixin


class VisitSequence:
    """A class that calculates the previous_visit and can enforce
    that visits are filled in sequence.
    """

    def __init__(self, appointment: Appointment, skip_enforce: bool | None = None) -> None:
        self._previous_appointment = None
        self.appointment = appointment
        self.skip_enforce = skip_enforce  # for tests
        self.appointment_model_cls = self.appointment.__class__
        descriptor = getattr(
            self.appointment_model_cls,
            self.appointment_model_cls.related_visit_model_attr(),
        )
        try:
            self.model_cls = descriptor.related.related_model
        except AttributeError:
            self.model_cls = descriptor.rel.related_model
        self.subject_identifier = self.appointment.subject_identifier
        self.visit_schedule_name = self.appointment.visit_schedule_name
        self.visit_code = self.appointment.visit_code
        self.visit_code_sequence = self.appointment.visit_code_sequence

    def enforce_sequence(self, document_type: str = None) -> None:
        """Raises an exception if sequence is not adhered to; that is,
        the visit reports are not completed in order.

        Do not enforce for SKIPPED appointments.
        """
        document_type = document_type or _("report")
        opts = dict(
            subject_identifier=self.appointment.subject_identifier,
            visit_schedule_name=self.appointment.visit_schedule_name,
            schedule_name=self.appointment.schedule_name,
            appt_datetime__lt=self.appointment.appt_datetime,
        )
        opts.update(**{f"{self.appointment.related_visit_model_attr()}__isnull": True})
        if (
            appointments := self.appointment.__class__.objects.filter(**opts)
            .exclude(appt_status=SKIPPED_APPT)
            .order_by("timepoint", "visit_code_sequence")
        ):
            msg = _(
                "Previous visit report required. Enter report for "
                "'%(visit_code)s.%(visit_code_sequence)s' "
                "before completing this %(document_type)s."
            ) % dict(
                visit_code=appointments.first().visit_code,
                visit_code_sequence=appointments.first().visit_code_sequence,
                document_type=document_type,
                appt=self.appointment,
            )
            raise VisitSequenceError(msg)

    @property
    def previous_visit_code(self) -> str:
        """Return the previous visit code or the existing
        visit code if sequence is not 0.
        """
        previous_visit_code = None
        if self.visit_code_sequence != 0:
            previous_visit_code = self.visit_code
        else:
            previous = self.appointment.schedule.visits.previous(self.visit_code)
            try:
                previous_visit_code = previous.code
            except AttributeError:
                pass
        return previous_visit_code

    @property
    def previous_appointment(self) -> Appointment | None:
        """Returns the previous appointment model instance or None.

        Considers interim appointments (relative_previous).
        """
        if not self._previous_appointment:
            if not self.skip_enforce:
                self.enforce_sequence()
            self._previous_appointment = self.appointment.relative_previous
        return self._previous_appointment

    @property
    def previous_visit(self) -> VisitModelMixin | None:
        """Returns the previous visit model instance if it exists
        or raises.

        A previous visit report must exist.
        """
        if self.previous_appointment:
            return self.model_cls.objects.get(appointment=self.previous_appointment)
        return None

    def get_previous_visit(self) -> VisitModelMixin | None:
        """Returns the previous visit model instance if it exists"""
        return self.previous_visit



from datetime import datetime
from typing import Protocol, TypeVar, Union

from django.db import models
from edc_model.stubs import ModelMetaStub


class RelatedVisitModelStub(Protocol):
    report_datetime: Union[datetime, models.DateTimeField]
    subject_identifier: Union[str, models.CharField]
    reason: str
    reason_unscheduled: str
    reason_unscheduled_other: str
    visit_code: Union[str, models.CharField]
    visit_code_sequence: Union[int, models.IntegerField]
    visit_schedule: Union[str, models.CharField]
    schedule: Union[str, models.CharField]
    study_status: str
    require_crfs: bool

    objects: models.Manager
    _meta: ModelMetaStub

    def natural_key(self) -> tuple: ...

    def save(self, *args, **kwargs) -> None: ...

    def related_visit_model_attr(self) -> str: ...

    def get_visit_reason_no_follow_up_choices(self) -> list: ...

    def get_reason_display(self) -> str: ...

    def get_reason_unscheduled_display(self) -> str: ...

    def get_require_crfs_display(self) -> str: ...

    def update_document_status_on_save(self, update_fields=None) -> None: ...


TRelatedVisitModelStub = TypeVar("TRelatedVisitModelStub", bound="RelatedVisitModelStub")



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Type, TypeVar

from django.apps import apps as django_apps
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

from .exceptions import RelatedVisitModelError

if TYPE_CHECKING:
    from edc_list_data.model_mixins import ListModelMixin

    from .models import SubjectVisit, SubjectVisitMissed

    ListModel = TypeVar("ListModel", bound=ListModelMixin)


def get_related_visit_model() -> str:
    """Returns the label_lower of the related visit model for this
    project.

    One `related visit model` allowed per project.
    """
    return getattr(settings, "SUBJECT_VISIT_MODEL", "edc_visit_tracking.subjectvisit")


def get_related_visit_model_cls() -> Type[SubjectVisit]:
    model_cls = django_apps.get_model(get_related_visit_model())
    if model_cls._meta.proxy:
        # raise for now until we have a solution
        raise RelatedVisitModelError(
            f"Not allowed. Related visit model may not be a proxy model. Got {model_cls}. "
        )
    return model_cls


def get_subject_visit_model() -> str:
    warnings.warn(
        "This func has been renamed to `get_related_visit_model`.",
        DeprecationWarning,
        stacklevel=2,
    )
    return get_related_visit_model()


def get_subject_visit_model_cls() -> Type[SubjectVisit]:
    warnings.warn(
        "This func has been renamed to `get_related_visit_model_cls`.",
        DeprecationWarning,
        stacklevel=2,
    )
    return get_related_visit_model_cls()


def get_subject_visit_missed_model() -> str:
    error_msg = (
        "Settings attribute `SUBJECT_VISIT_MISSED_MODEL` not set. Update settings. "
        "For example, `SUBJECT_VISIT_MISSED_MODEL=meta_subject.subjectvisitmissed`. "
        "See also `SubjectVisitMissedModelMixin`."
    )
    try:
        model = settings.SUBJECT_VISIT_MISSED_MODEL
    except AttributeError as e:
        raise ImproperlyConfigured(f"{error_msg} Got {e}.")
    else:
        if not model:
            raise ImproperlyConfigured(f"{error_msg} Got None.")
    return model


def get_allow_missed_unscheduled_appts() -> bool:
    """Returns value of settings attr or False"""
    return getattr(settings, "EDC_VISIT_TRACKING_ALLOW_MISSED_UNSCHEDULED", False)


def get_subject_visit_missed_model_cls() -> Type[SubjectVisitMissed]:
    return django_apps.get_model(get_subject_visit_missed_model())


def get_previous_related_visit(
    related_visit: SubjectVisit, include_interim=None
) -> SubjectVisit | None:
    if related_visit:
        if include_interim:
            previous_appointment = related_visit.appointment.relative_previous
        else:
            previous_appointment = related_visit.appointment.previous
        return getattr(previous_appointment, "related_visit", None)
    return None


def get_next_related_visit(
    related_visit: SubjectVisit, include_interim=None
) -> SubjectVisit | None:
    if related_visit:
        if include_interim:
            next_appointment = related_visit.appointment.relative_next
        else:
            next_appointment = related_visit.appointment.next
        return getattr(next_appointment, "related_visit", None)
    return None



from edc_consent.consent_definition import ConsentDefinition
from edc_protocol.research_protocol_config import ResearchProtocolConfig

consent_v1 = ConsentDefinition(
    "visit_tracking_app.subjectconsentv1",
    start=ResearchProtocolConfig().study_open_datetime,
    end=ResearchProtocolConfig().study_close_datetime,
    gender=["M", "F"],
    version="1",
    age_min=16,
    age_max=64,
    age_is_adult=18,
)



from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist
from edc_utils import get_utcnow
from model_bakery import baker

from ..constants import SCHEDULED


class VisitTestHelperFutureDate(Exception):
    pass


class VisitTestHelperError(Exception):
    pass


class VisitTestHelper:
    appointment_model = "edc_appointment.appointment"

    @property
    def appointment_model_cls(self):
        return django_apps.get_model(self.appointment_model)

    def add_visit(
        self, model_label=None, visit_code=None, reason=None, subject_identifier=None
    ):
        """Adds (or gets) and returns a visit for give model and code."""
        reason = reason or SCHEDULED
        model = django_apps.get_model(*model_label.split("."))
        try:
            appointment = self.appointment_model_cls.objects.get(
                subject_identifier=subject_identifier, visit_code=visit_code
            )
            if appointment.appt_datetime > get_utcnow():
                raise VisitTestHelperFutureDate(
                    f"For testing, create visits without future dates. "
                    f"Got {appointment.visit_code}, {appointment.appt_datetime}"
                )
            try:
                visit = self.get_visit(
                    visit_code=visit_code,
                    model_label=model_label,
                    subject_identifier=subject_identifier,
                )
            except model.DoesNotExist:
                # create visit for the first time
                visit = baker.make_recipe(
                    model_label,
                    appointment=appointment,
                    report_datetime=appointment.appt_datetime,
                    reason=reason,
                )
        except ObjectDoesNotExist as e:
            raise VisitTestHelperError(
                f"{e} {model_label}, {visit_code}. Did you complete the enrollment form?"
            )
        return visit

    def add_visits(self, *codes, model_label=None, subject_identifier=None, reason=None):
        """Adds a sequence of visits for the codes provided.

        If an infant visit already exists, it will just pass.
        """
        visits = []
        for code in codes:
            visits.append(
                self.add_visit(
                    model_label=model_label,
                    visit_code=code,
                    reason=reason,
                    subject_identifier=subject_identifier,
                )
            )
        return visits

    def get_visit(self, visit_code=None, model_label=None, subject_identifier=None):
        """Returns a visit instance if it exists."""
        model = django_apps.get_model(*model_label.split("."))
        visit = model.objects.get(
            appointment__subject_identifier=subject_identifier, visit_code=visit_code
        )
        return visit

    def get_last_visit(self, model_label=None, subject_identifier=None):
        """Returns the last visit instance if it exists."""
        model = django_apps.get_model(*model_label.split("."))
        return (
            model.objects.filter(appointment__subject_identifier=subject_identifier)
            .order_by("report_datetime")
            .last()
        )



from django.core.validators import MinValueValidator
from django.db import models
from django.utils.translation import gettext_lazy as _
from edc_constants.choices import ALIVE_DEAD_UNKNOWN, YES_NO, YES_NO_NA
from edc_constants.constants import NO, NOT_APPLICABLE
from edc_model import models as edc_models
from edc_model.validators import date_not_future
from edc_protocol.validators import date_not_before_study_start

from ..constants import VISIT_MISSED_ACTION


class SubjectVisitMissedModelMixin(models.Model):
    """Declare with:

        missed_reasons = models.ManyToManyField(SubjectVisitMissedReasons, blank=True)

    And include in your `lists` app:

        class SubjectVisitMissed(
            CrfModelMixin,
            SubjectVisitMissedModelMixin,
            edc_models.BaseUuidModel):

            subject_visit = models.OneToOneField(
                settings.SUBJECT_VISIT_MODEL,
                on_delete=models.PROTECT,
            )

            missed_reasons = models.ManyToManyField(
                SubjectVisitMissedReasons, blank=True
            )

            class Meta(CrfModelMixin.Meta, edc_models.BaseUuidModel.Meta):
                verbose_name = "Missed Visit Report"
                verbose_name_plural = "Missed Visit Report"
    """

    action_name = VISIT_MISSED_ACTION

    survival_status = models.CharField(
        verbose_name=_("Survival status"),
        max_length=25,
        choices=ALIVE_DEAD_UNKNOWN,
        help_text=_("If deceased, complete the death report"),
    )

    contact_attempted = models.CharField(
        verbose_name=_(
            "Were any attempts made to contact the participant "
            "since the expected appointment date?"
        ),
        max_length=25,
        choices=YES_NO,
        help_text=_("Not including pre-appointment reminders"),
    )

    contact_attempts_count = models.IntegerField(
        verbose_name=_(
            "Number of attempts made to contact participant"
            "since the expected appointment date"
        ),
        validators=[MinValueValidator(1)],
        help_text=_(
            "Not including pre-appointment reminders. Multiple attempts "
            "on the same day count as a single attempt."
        ),
        null=True,
        blank=True,
    )

    contact_attempts_explained = models.TextField(
        verbose_name=_("If contact not made and less than 3 attempts, please explain"),
        null=True,
        blank=True,
    )

    contact_last_date = models.DateField(
        verbose_name=_("Date of last telephone contact/attempt"),
        validators=[date_not_future, date_not_before_study_start],
        null=True,
        blank=True,
    )

    contact_made = models.CharField(
        verbose_name=_("Was contact finally made with the participant?"),
        max_length=25,
        choices=YES_NO_NA,
        default=NOT_APPLICABLE,
    )

    missed_reasons = models.ManyToManyField(
        "edc_visit_tracking.SubjectVisitMissedReasons", blank=True
    )

    missed_reasons_other = edc_models.OtherCharField()

    ltfu = models.CharField(
        verbose_name=_("Has the participant met the protocol criteria for lost to follow up?"),
        max_length=15,
        choices=YES_NO_NA,
        default=NO,
        help_text=_("If 'Yes', complete the Loss to Follow up form"),
    )

    comment = models.TextField(
        verbose_name=_("Please provide further details, if any"),
        null=True,
        blank=True,
    )

    class Meta:
        abstract = True
        indexes = [models.Index(fields=["action_identifier", "site", "id"])]



from datetime import date

from django.db import models
from django.db.models.deletion import PROTECT
from edc_consent.managers import ConsentObjectsByCdefManager, CurrentSiteByCdefManager
from edc_constants.constants import HOSPITALIZED, OTHER
from edc_crf.model_mixins import CrfInlineModelMixin
from edc_identifier.managers import SubjectIdentifierManager
from edc_identifier.model_mixins import NonUniqueSubjectIdentifierFieldMixin
from edc_lab.model_mixins import RequisitionModelMixin
from edc_list_data.model_mixins import ListModelMixin
from edc_model.models import BaseUuidModel
from edc_offstudy.model_mixins import OffstudyModelMixin
from edc_registration.model_mixins import UpdatesOrCreatesRegistrationModelMixin
from edc_screening.model_mixins import ScreeningModelMixin
from edc_sites.model_mixins import SiteModelMixin
from edc_utils import get_utcnow
from edc_visit_schedule.model_mixins import OffScheduleModelMixin, OnScheduleModelMixin

from edc_visit_tracking.model_mixins import VisitTrackingCrfModelMixin
from edc_visit_tracking.models import SubjectVisit

from .consents import consent_v1

list_data = {
    "edc_visit_tracking.subjectvisitmissedreasons": [
        ("forgot", "Forgot / Can’t remember being told about appointment"),
        ("family_emergency", "Family emergency (e.g. funeral) and was away"),
        ("travelling", "Away travelling/visiting"),
        ("working_schooling", "Away working/schooling"),
        ("too_sick", "Too sick or weak to come to the centre"),
        ("lack_of_transport", "Transportation difficulty"),
        (HOSPITALIZED, "Hospitalized"),
        (OTHER, "Other reason (specify below)"),
    ],
}


class SubjectScreening(ScreeningModelMixin, BaseUuidModel):
    consent_definition = consent_v1
    objects = SubjectIdentifierManager()


class SubjectConsent(
    SiteModelMixin,
    NonUniqueSubjectIdentifierFieldMixin,
    UpdatesOrCreatesRegistrationModelMixin,
    BaseUuidModel,
):
    report_datetime = models.DateTimeField(default=get_utcnow)

    consent_datetime = models.DateTimeField(default=get_utcnow)

    version = models.CharField(max_length=25, default="1")

    identity = models.CharField(max_length=25)

    confirm_identity = models.CharField(max_length=25)

    dob = models.DateField(default=date(1995, 1, 1))


class SubjectConsentV1(SubjectConsent):
    objects = ConsentObjectsByCdefManager()
    on_site = CurrentSiteByCdefManager()

    class Meta:
        proxy = True


class OnScheduleOne(SiteModelMixin, OnScheduleModelMixin, BaseUuidModel):
    class Meta(BaseUuidModel.Meta):
        pass


class OnScheduleTwo(SiteModelMixin, OnScheduleModelMixin, BaseUuidModel):
    class Meta(BaseUuidModel.Meta):
        pass


class OffSchedule(SiteModelMixin, OffScheduleModelMixin, BaseUuidModel):
    class Meta(BaseUuidModel.Meta):
        pass


class OffScheduleOne(SiteModelMixin, OffScheduleModelMixin, BaseUuidModel):
    class Meta(BaseUuidModel.Meta):
        pass


class SubjectOffstudy(SiteModelMixin, OffstudyModelMixin, BaseUuidModel):
    class Meta(OffstudyModelMixin.Meta):
        pass


class SubjectRequisition(RequisitionModelMixin, BaseUuidModel):
    def update_reference_on_save(self):
        pass

    class Meta(RequisitionModelMixin.Meta):
        pass


class CrfOne(SiteModelMixin, VisitTrackingCrfModelMixin, BaseUuidModel):
    f1 = models.CharField(max_length=50, null=True)

    f2 = models.CharField(max_length=50, null=True)

    f3 = models.CharField(max_length=50, null=True)

    class Meta(BaseUuidModel.Meta):
        pass


class CrfTwo(SiteModelMixin, VisitTrackingCrfModelMixin, BaseUuidModel):
    f1 = models.CharField(max_length=50, null=True)

    f2 = models.CharField(max_length=50, null=True)

    f3 = models.CharField(max_length=50, null=True)

    class Meta(BaseUuidModel.Meta):
        pass


class CrfThree(SiteModelMixin, VisitTrackingCrfModelMixin, BaseUuidModel):
    f1 = models.CharField(max_length=50, null=True)

    f2 = models.CharField(max_length=50, null=True)

    f3 = models.CharField(max_length=50, null=True)

    class Meta(BaseUuidModel.Meta):
        pass


class CrfFour(SiteModelMixin, VisitTrackingCrfModelMixin, BaseUuidModel):
    f1 = models.CharField(max_length=50, null=True)

    f2 = models.CharField(max_length=50, null=True)

    f3 = models.CharField(max_length=50, null=True)

    class Meta(BaseUuidModel.Meta):
        pass


class CrfFive(SiteModelMixin, VisitTrackingCrfModelMixin, BaseUuidModel):
    f1 = models.CharField(max_length=50, null=True)

    f2 = models.CharField(max_length=50, null=True)

    f3 = models.CharField(max_length=50, null=True)

    class Meta(BaseUuidModel.Meta):
        pass


class BadCrfNoRelatedVisit(SiteModelMixin, VisitTrackingCrfModelMixin, BaseUuidModel):
    subject_visit = None

    f1 = models.CharField(max_length=50, null=True)

    f2 = models.CharField(max_length=50, null=True)

    f3 = models.CharField(max_length=50, null=True)

    class Meta(BaseUuidModel.Meta):
        pass


class OtherModel(SiteModelMixin, BaseUuidModel):
    f1 = models.CharField(max_length=10, default="erik")

    class Meta(BaseUuidModel.Meta):
        pass


class CrfOneInline(CrfInlineModelMixin, BaseUuidModel):
    crf_one = models.ForeignKey(CrfOne, on_delete=PROTECT)

    other_model = models.ForeignKey(OtherModel, on_delete=PROTECT)

    f1 = models.CharField(max_length=10, default="erik")

    def natural_key(self) -> tuple:
        return tuple()

    class Meta(CrfInlineModelMixin.Meta):
        crf_inline_parent = "crf_one"


class BadCrfOneInline(CrfInlineModelMixin, BaseUuidModel):
    """A model class missing _meta.crf_inline_parent."""

    crf_one = models.ForeignKey(CrfOne, on_delete=PROTECT)

    other_model = models.ForeignKey(OtherModel, on_delete=PROTECT)

    f1 = models.CharField(max_length=10, default="erik")

    def natural_key(self) -> tuple:
        return tuple()

    class Meta:
        pass


class BadCrfOneInline2(CrfInlineModelMixin, BaseUuidModel):
    crf_one = models.ForeignKey(CrfOne, on_delete=PROTECT)

    other_model = models.ForeignKey(OtherModel, on_delete=PROTECT)

    f1 = models.CharField(max_length=10, default="erik")

    def natural_key(self) -> tuple:
        return tuple()

    class Meta(CrfInlineModelMixin.Meta):
        crf_inline_parent = None


class CustomSubjectVisitMissedReasons(ListModelMixin):
    class Meta(ListModelMixin.Meta):
        verbose_name = "Subject Visit Missed Reason"
        verbose_name_plural = "Subject Visit Missed Reasons"


class SubjectVisit2(SubjectVisit):
    class Meta:
        proxy = True
